#ifndef KOLIBRIX_SYSCALLS_H_INCLUDED
#define KOLIBRIX_SYSCALLS_H_INCLUDED

#include <stddef.h>

#ifndef _NORETURN_
#  define _NORETURN_ __attribute__((noreturn))
#endif

#pragma pack(push, 1)
struct ThreadInfo {
  uint32_t usage;
  uint16_t win_pos;
  uint16_t win_slot;
  uint16_t reserved0;
  char     name[11];
  char     reserved1;
  void     *adr;
  uint32_t mem_used;
  uint32_t pid;
  int32_t  win_x;
  int32_t  win_y;
  int32_t  win_w;
  int32_t  win_h;
  uint16_t state;
  uint16_t reserved2;
  int32_t  client_x;
  int32_t  client_y;
  int32_t  client_w;
  int32_t  client_h;
  uint8_t  win_state;
  uint32_t events;
  uint8_t  kbd_mode;
  char     reserved[948];
};

struct FileExt {
  int fn;
  int a, b, c, d;
  int enc;
  const void *path;
};

struct FileTime {
  char s, m, h, reserved;
};

struct FileDate {
  char d, m;
  unsigned short y;
};

struct FileInfo {
  int attr;
  int enc;
  struct FileTime ctime;
  struct FileDate cdate;
  struct FileTime atime;
  struct FileDate cdate;
  struct FileTime mtime;
  struct FileDate mdate;
  long long size;
  union {
    char  cp866[264];
    short utf16[260];
    char  utf8[520];
  };
};

struct IPC_Buffer {
  int lock;
  int size;
  char data[];
};

struct IPC_Message {
  int pid;
  int size;
  char data[];
};

typedef struct DebugBuffer {
  int size; /* without this header. >0: unlocked, <0: locked */
  int used;
  char data[];
} DebugBuffer;

typedef struct DebugCommon {
  int type; /* DEBUG_TERMINATED */
  int pid;
} DebugCommon;

typedef struct DebugException {
  int type; /* DEBUG_EXCEPTION, DEBUG_BREAKPOINT */
  int pid;
  int code;
} DebugException;

typedef union DebugMessage {
  int type;
  struct DebugCommon common;
  struct DebugException exception;
} DebugMessage;

typedef struct DebugContext {
  int eip;
  int eflags;
  int eax;
  int ecx;
  int edx;
  int ebx;
  int esp;
  int ebp;
  int esi;
  int edi;
  /* --- extended context --- */
  /* TODO */
} DebugContext;
#pragma pack(pop)

#define ENC_DEFAULT 0
#define ENC_CP866   1
#define ENC_UTF16   2
#define ENC_UTF8    3

/*****************************************************************************/
/**                          file access management                         **/
/*****************************************************************************/

#define FILE_ATTR_RDONLY (1 << 0)
#define FILE_ATTR_HIDDEN (1 << 1)
#define FILE_ATTR_SYSTEM (1 << 2)
#define FILE_ATTR_VOLUME (1 << 3)
#define FILE_ATTR_FOLDER (1 << 4)
#define FILE_ATTR_ARCHIVE (1 << 5)

#define FILE_SUCCESS       0
#define FILE_NOT_SUPPORTED 2
#define FILE_UNKNOWN_FS    3
#define FILE_NOT_FOUND     5
#define FILE_EOF           6
#define FILE_INVALID_PTR   7
#define FILE_FULL_DISK     8
#define FILE_FS_ERROR      9
#define FILE_ACCESS_DENIED 10
#define FILE_DEVICE_ERROR  11
#define FILE_FS_NOMEM      12
#define FILE_APP_OUT_OF_MEMORY 30
#define FILE_APP_NOT_EXE   31
#define FILE_APP_TOO_MANY_PROCESS 32

#define APP_FLAG_DEBUG 1

static inline int FileExt (struct FileExt *info, int *ret) {
  int ret1, ret2;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (ret1),
      "=b" (ret2)
    : "a" (80),
      "b" (info)
    : "memory"
  );
  if (ret) *ret = ret2;
  return ret1;
}

static inline int ReadFile (int offset, void *data, size_t len, int enc, const void *path, int *count) {
  struct FileExt info = { 0, offset, 0/*high_offset*/, (int)len, (int)data, enc, path };
  return FileExt(&info, count);
}

static inline int CreateFile (const void *data, size_t len, int enc, const void *path, int *count) {
  struct FileExt info = { 2, 0, 0, (int)len, (int)data, enc, path };
  return FileExt(&info, count);
}

static inline int WriteFile (int offset, const void *data, size_t len, int enc, const void *path, int *count) {
  struct FileExt info = { 3, offset, 0/*high_offset*/, (int)len, (int)data, enc, path };
  return FileExt(&info, count);
}

static inline int InfoFile (int flags, struct FileInfo *bdfe, int enc, const void *path) {
  struct FileExt info = { 5, 0, flags, 0, (int)bdfe, enc, path };
  return FileExt(&info, NULL);
}

static inline int StartApp (int flags, const char *params, int enc, const void *path) {
  struct FileExt info = { 7, flags, (int)params, 0, 0, enc, path };
  return FileExt(&info, NULL);
}

static inline int DeleteFile (int enc, const void *path) {
  struct FileExt info = { 8, 0, 0, 0, 0, enc, path };
  return FileExt(&info, NULL);
}

/*****************************************************************************/
/**                             board functions                             **/
/*****************************************************************************/

static inline BoardWrite (char ch) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (63),
      "b" (1),
      "c" (ch)
  );
}

/*****************************************************************************/
/**                             event functions                             **/
/*****************************************************************************/

#define EVENT_NONE     0
#define EVENT_REDRAW   1
#define EVENT_KEYBOARD 2
#define EVENT_BUTTON   3
#define EVENT_RESERVED 4
#define EVENT_REDRAWBG 5
#define EVENT_MOUSE    6
#define EVENT_IPC      7
#define EVENT_NETWORK  8
#define EVENT_DEBUG    9
#define EVENT_IRQ0     16
#define EVENT_IRQ1     17
#define EVENT_IRQ2     18
#define EVENT_IRQ3     19
#define EVENT_IRQ4     20
#define EVENT_IRQ5     21
#define EVENT_IRQ6     22
#define EVENT_IRQ7     23
#define EVENT_IRQ8     24
#define EVENT_IRQ9     25
#define EVENT_IRQ10    26
#define EVENT_IRQ11    27
#define EVENT_IRQ12    28
#define EVENT_IRQ13    29
#define EVENT_IRQ14    30
#define EVENT_IRQ15    31

#define EVENT_REDRAW_BIT        0
#define EVENT_KEYBOARD_BIT      1
#define EVENT_BUTTON_BIT        2
#define EVENT_RESERVED_BIT      3
#define EVENT_REDRAWBG_BIT      4
#define EVENT_MOUSE_BIT         5
#define EVENT_IPC_BIT           6
#define EVENT_NETWORK_BIT       7
#define EVENT_DEBUG_BIT         8
#define EVENT_CURSOR_FILTER_BIT 30
#define EVENT_MOUSE_FILTER_BIT  31

#define EVENT_REDRAW_FLAG        (1 << EVENT_REDRAW_BIT)
#define EVENT_KEYBOARD_FLAG      (1 << EVENT_KEYBOARD_BIT)
#define EVENT_BUTTON_FLAG        (1 << EVENT_BUTTON_BIT)
#define EVENT_RESERVED_FLAG      (1 << EVENT_RESERVED_BIT)
#define EVENT_REDRAWBG_FLAG      (1 << EVENT_REDRAWBG_BIT)
#define EVENT_MOUSE_FLAG         (1 << EVENT_MOUSE_BIT)
#define EVENT_IPC_FLAG           (1 << EVENT_IPC_BIT)
#define EVENT_NETWORK_FLAG       (1 << EVENT_NETWORK_BIT)
#define EVENT_DEBUG_FLAG         (1 << EVENT_DEBUG_BIT)
#define EVENT_CURSOR_FILTER_FLAG (1 << EVENT_CURSOR_FILTER_BIT)
#define EVENT_MOUSE_FILTER_FLAG  (1 << EVENT_MOUSE_FILTER_BIT)

#define EVENT_DEFAULT_MASK (EVENT_REDRAW_FLAG | EVENT_KEYBOARD_FLAG | EVENT_BUTTON_FLAG)

static inline int SetEventMask (int mask) {
  int ret;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (ret)
    : "a" (40),
      "b" (mask)
    : "memory"
  );
  return ret;
}

static inline int WaitEvent (void) {
  int ret;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (ret)
    : "a" (10)
    : "memory"
  );
  return ret;
}

static inline int CheckEvent (void) {
  int ret;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (ret)
    : "a" (11)
    : "memory"
  );
  return ret;
}

static inline int WaitEventTimeout (unsigned t) {
  int ret;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (ret)
    : "a" (23),
      "b" (t) // 1/100 sec
    : "memory"
  );
  return ret;
}


/*****************************************************************************/
/**                            memory management                            **/
/*****************************************************************************/

static inline int MemoryResize (uintptr_t size) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (64),
      "b" (1),
      "c" (size)
    : "memory"
  );
  return res;
}

/*****************************************************************************/
/**                        advanced memory management                       **/
/*****************************************************************************/

static inline int HeapInit (void) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (68),
      "b" (11)
    : "memory"
  );
  return res;
}

static inline int MemAlloc (int size) {
  void *res;
  __asm__ __volatile__ (
    "int $0x40" 
    : "=a" (res)
    : "a" (68),
      "b" (12),
      "c" (size)
    : "memory"
  );
  return res;
}

static inline int MemFree (void *ptr) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (68),
      "b" (13),
      "c" (ptr)
    : "memory"
  );
  return res;
}

static inline void *MemRealloc (int size, void *ptr) {
  void *res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (68),
      "b" (20),
      "c" (size),
      "d" (ptr)
    : "memory"
  );
  return res;
}

static inline void MemFreeExt (void *ptr, int offset, int size) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (68),
      "b" (26),
      "c" (ptr),
      "d" (offset),
      "S" (size)
    : "memory"
  );
}

static inline void *LoadFileExt (void *path, int enc, int *size) {
  void *res;
  int sz;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res),
      "=d" (sz)
    : "a" (68),
      "b" (28),
      "c" (path),
      "d" (enc)
    : "memory"
  );
  if (size) {
    *size = sz;
  }
  return res;
}

static inline int MemAllocRing (int size) {
  void *res;
  __asm__ __volatile__ (
    "int $0x40" 
    : "=a" (res)
    : "a" (68),
      "b" (29),
      "c" (size)
    : "memory"
  );
  return res;
}

/*****************************************************************************/
/**                             process control                             **/
/*****************************************************************************/

/* ThreadInfo.state */
#define THREAD_RUNNING     0
#define THREAD_SUSPENDED   1
#define THREAD_SUSPENDWAIT 2
#define THREAD_TERMINATED  3
#define THREAD_CRASHED     4
#define THREAD_WAIT        5
#define THREAD_FREE        9

static void Delay (unsigned t) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (5),
      "b" (t) /* 1/100 secs */
  );
}

static inline int GetThreadInfo (struct ThreadInfo *info, int slot) {
  int ret;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (ret)
    : "a" (9),
      "b" (info),
      "c" (slot)
    : "memory"
  );
  return ret;
}

static inline int GetThreadSlot (int pid) {
  int slot;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (slot)
    : "a" (18),
      "b" (21),
      "c" (pid)
    : "memory"
  );
  return slot;
}

static inline int CreateThread (void *eip, void *esp) {
  int tid;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (tid)
    : "a" (51),
      "b" (1),
      "c" (eip),
      "d" (esp)
    : "memory"
  );
  return tid;
}

static inline void Terminate (void) _NORETURN_ {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (-1)
    : "memory"
  );
}

static inline int TerminateThreadID (int pid) _NORETURN_ {
  int ret;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (ret)
    : "a" (18),
      "b" (18),
      "c" (pid)
    : "memory"
  );
  return ret;
}

static inline void SwitchTask (void) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (68),
      "b" (1)
  );
}

/*****************************************************************************/
/**                             work directory                              **/
/*****************************************************************************/

static inline void SetCWDExt (void *cwd, int enc) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (30),
      "b" (4),
      "c" (cwd),
      "d" (enc)
    : "memory"
  );
}

static inline int GetCWDExt (void *buf, int size, int enc) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (30),
      "b" (5),
      "c" (buf),
      "d" (size),
      "S" (enc)
    : "memory"
  );
  return res;
}

/*****************************************************************************/
/**                             synchronization                             **/
/*****************************************************************************/

static inline int CreateFutex (int *var) {
  int futex;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (futex)
    : "a" (77),
      "b" (0),
      "c" (var)
    : "memory"
  );
  return futex;
}

static inline int DestroyFutex (int futex) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (77),
      "b" (1),
      "c" (futex)
    : "memory"
  );
  return res;
}

static inline int WaitFutex (int futex, int value, unsigned timeout) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (77),
      "b" (2),
      "c" (futex),
      "d" (value),
      "S" (timeout)
    : "memory"
  );
  return res;
}

static inline int WakeFutex (int futex, int n) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (77),
      "b" (3),
      "c" (futex),
      "d" (n)
    : "memory"
  );
  return res;
}

/*****************************************************************************/
/**                       inter process communication                       **/
/*****************************************************************************/

/* SendMessage codes */
#define IPC_SUCCESS  0
#define IPC_NOBUFFER 1
#define IPC_LOCKED   2
#define IPC_OWERFLOW 3
#define IPC_DEAD     4

static inline int SetArea (void *buf, int size) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (60),
      "b" (1),
      "c" (buf),
      "d" (size)
    : "memory"
  );
  return res;
}

static inline int SendMessage (int pid, void *data, int size) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (60),
      "b" (2),
      "c" (pid),
      "d" (data),
      "S" (size)
    : "memory"
  );
  return res;
}

/*****************************************************************************/
/**                                debug api                                **/
/*****************************************************************************/

#define DEBUG_EXCEPTION  1
#define DEBUG_TERMINATED 2
#define DEBUG_BREAKPOINT 3

#define DEBUG_FLAG_SET       (1 << 7)

#define DEBUG_FLAG_EXCEPTION (0 << 0)
#define DEBUG_FLAG_READ      (1 << 0)
#define DEBUG_FLAG_READWRITE (2 << 0)

#define DEBUG_FLAG_BYTE      (0 << 2)
#define DEBUG_FLAG_WORD      (1 << 2)
#define DEBUG_FLAG_DWORD     (2 << 2)

static inline void DebugSetMessageArea (void *ptr) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (69),
      "b" (0),
      "c" (buf)
    : "memory"
  );
}

static inline void DebugGetRegisters (int pid, int size, void *ptr) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (69),
      "b" (1),
      "c" (pid),
      "d" (size),
      "S" (ptr)
    : "memory"
  );
}

static inline void DebugSetRegisters (int pid, int size, void *ptr) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (69),
      "b" (2),
      "c" (pid),
      "d" (size),
      "S" (ptr)
    : "memory"
  );
}

static inline void DebugDetach (int pid) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (69),
      "b" (3),
      "c" (pid)
    : "memory"
  );
}

static inline void DebugSuspend (int pid) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (69),
      "b" (4),
      "c" (pid)
    : "memory"
  );
}

static inline void DebugResume (int pid) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (69),
      "b" (5),
      "c" (pid)
    : "memory"
  );
}

static inline int DebugReadMemory (int pid, int len, int adr, void *buf) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (69),
      "b" (6),
      "c" (pid),
      "d" (len),
      "S" (adr),
      "D" (buf)
    : "memory"
  );
  return res;
}

static inline int DebugWriteMemory (int pid, int len, int adr, void *buf) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (69),
      "b" (7),
      "c" (pid),
      "d" (len),
      "S" (adr),
      "D" (buf)
    : "memory"
  );
  return res;
}

static inline DebugTerminate (int pid) {
  __asm__ __volatile__ (
    "int $0x40"
    :
    : "a" (69),
      "b" (8),
      "c" (pid)
    : "memory"
  );
}

static inline int DebugBreakpoint (int pid, int index, int flags, int adr) {
  int res;
  __asm__ __volatile__ (
    "int $0x40"
    : "=a" (res)
    : "a" (69),
      "b" (8),
      "c" (pid),
      "d" (flags | index << 8),
      "S" (adr)
    : "memory"
  );
  return res;
}

#endif /* KOLIBRIX_SYSCALLS_H_INCLUDED */
